"""
API for ytData frontend




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2014, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from .data_structures import \
    YTDataContainerDataset, \
    YTSpatialPlotDataset, \
    YTGridDataset, \
    YTGridHierarchy, \
    YTGrid, \
    YTNonspatialDataset, \
    YTNonspatialHierarchy, \
    YTNonspatialGrid, \
    YTProfileDataset, \
    YTClumpContainer, \
    YTClumpTreeDataset

from .io import \
    IOHandlerYTDataContainerHDF5, \
    IOHandlerYTGridHDF5, \
    IOHandlerYTSpatialPlotHDF5, \
    IOHandlerYTNonspatialhdf5

from .fields import \
    YTDataContainerFieldInfo, \
    YTGridFieldInfo

from .utilities import \
    save_as_dataset

from . import tests
